# Copyright (C) 2020 by ÿnérant, eichhornchen, nicomarg, charlse
# SPDX-License-Identifier: GPL-3.0-or-later

from random import choice, randint
from typing import Optional

from .player import Player
from ..interfaces import Entity, FightingEntity, Map
from ..translations import gettext as _


class Item(Entity):
    """
    A class for items
    """
    held: bool
    held_by: Optional[Player]

    def __init__(self, held: bool = False, held_by: Optional[Player] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.held = held
        self.held_by = held_by

    def drop(self) -> None:
        """
        The item is dropped from the inventory onto the floor
        """
        if self.held:
            self.held_by.inventory.remove(self)
            self.map.add_entity(self)
            self.move(self.held_by.y, self.held_by.x)
            self.held = False
            self.held_by = None

    def use(self) -> None:
        """
        Indicates what should be done when the item is used.
        """

    def equip(self) -> None:
        """
        Indicates what should be done when the item is equipped.
        """

    def hold(self, player: "Player") -> None:
        """
        The item is taken from the floor and put into the inventory
        """
        self.held = True
        self.held_by = player
        self.map.remove_entity(self)
        player.inventory.append(self)

    def save_state(self) -> dict:
        """
        Saves the state of the entity into a dictionary
        """
        d = super().save_state()
        d["held"] = self.held
        return d


class Heart(Item):
    """
    A heart item to return health to the player
    """
    healing: int

    def __init__(self, name: str = "heart", healing: int = 5, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.healing = healing

    def hold(self, player: "Player") -> None:
        """
        When holding a heart, heal the player and don't put item in inventory.
        """
        player.health = min(player.maxhealth, player.health + self.healing)
        self.map.remove_entity(self)

    def save_state(self) -> dict:
        """
        Saves the state of the header into a dictionary
        """
        d = super().save_state()
        d["healing"] = self.healing
        return d


class Bomb(Item):
    """
    A bomb item intended to deal damage to enemies at long range
    """
    damage: int = 5
    exploding: bool
    owner: Optional["Player"]
    tick: int

    def __init__(self, name: str = "bomb", damage: int = 5,
                 exploding: bool = False, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.damage = damage
        self.exploding = exploding
        self.tick = 4
        self.owner = None

    def use(self) -> None:
        """
        When the bomb is used, throw it and explodes it.
        """
        if self.held:
            self.owner = self.held_by
            super().drop()
            self.exploding = True

    def act(self, m: Map) -> None:
        """
        Special exploding action of the bomb
        """
        if self.exploding:
            if self.tick > 0:
                # The bomb will explode in <tick> moves
                self.tick -= 1
            else:
                # The bomb is exploding.
                # Each entity that is close to the bomb takes damages.
                # The player earn XP if the entity was killed.
                log_message = _("Bomb is exploding.")
                for e in m.entities.copy():
                    if abs(e.x - self.x) + abs(e.y - self.y) <= 3 and \
                            isinstance(e, FightingEntity):
                        log_message += " " + e.take_damage(self, self.damage)
                        if e.dead:
                            self.owner.add_xp(randint(3, 7))
                m.logs.add_message(log_message)
                m.entities.remove(self)

    def save_state(self) -> dict:
        """
        Saves the state of the bomb into a dictionary
        """
        d = super().save_state()
        d["exploding"] = self.exploding
        d["damage"] = self.damage
        return d


class BodySnatchPotion(Item):
    """
    The body-snatch potion allows to exchange all characteristics with a random
    other entity.
    """

    def __init__(self, name: str = "body_snatch_potion", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def use(self) -> None:
        """
        Find a valid random entity, then exchange characteristics.
        """
        valid_entities = self.held_by.map.find_entities(FightingEntity)
        valid_entities.remove(self.held_by)
        entity = choice(valid_entities)
        entity_state = entity.save_state()
        player_state = self.held_by.save_state()
        self.held_by.__dict__.update(entity_state)
        entity.__dict__.update(player_state)
        self.held_by.map.currenty, self.held_by.map.currentx = self.held_by.y,\
            self.held_by.x

        self.held_by.inventory.remove(self)
